From dc2425ceaa8dd4fe68d1b1dc780de15800041564 Mon Sep 17 00:00:00 2001
From: Jacob Champion <jchampion@timescale.com>
Date: Mon, 26 Sep 2022 13:23:51 -0700
Subject: [PATCH v4 2/2] WIP: introduce publish_via_parent for logical
 replication

This new relation option allows regular inherited tables to be published
via their root table, just like partitions. This works by hijacking
pg_inherit's inhseqno column, and replacing a (single) existing entry
for the child with the value zero, indicating that it should be treated
as a logical partition by the publication machinery.

Initial sync works by asking the publisher for a list of logical
descendants of the published table, then COPYing them one-by-one into
the root. The publisher reuses the existing pubviaroot logic, adding the
new logical roots to code that previously looked only for partition
roots.

Known bugs/TODOs:
- I haven't given any thought to interactions with row filters, or to
  column lists.
- I'm not sure that I'm taking all the necessary locks yet, and those I
  do take may be taken in the wrong order.
---
 src/backend/access/common/reloptions.c      |  11 +
 src/backend/catalog/pg_inherits.c           | 123 ++++++-
 src/backend/catalog/pg_publication.c        | 336 +++++++++++++-----
 src/backend/commands/publicationcmds.c      |  12 +
 src/backend/commands/tablecmds.c            | 161 ++++++++-
 src/backend/partitioning/partdesc.c         |   3 +-
 src/backend/replication/logical/tablesync.c | 257 +++++++++-----
 src/backend/replication/pgoutput/pgoutput.c |   1 -
 src/bin/pg_dump/t/002_pg_dump.pl            |  24 ++
 src/bin/psql/tab-complete.c                 |   1 +
 src/include/catalog/pg_inherits.h           |   8 +-
 src/include/catalog/pg_proc.dat             |   7 +
 src/include/utils/rel.h                     |  12 +
 src/test/regress/expected/publication.out   | 359 ++++++++++++++++++++
 src/test/regress/sql/publication.sql        | 166 +++++++++
 src/test/subscription/t/013_partition.pl    | 312 +++++++++++++++++
 16 files changed, 1613 insertions(+), 180 deletions(-)

diff --git a/src/backend/access/common/reloptions.c b/src/backend/access/common/reloptions.c
index 469de9bb49..1ef081b78d 100644
--- a/src/backend/access/common/reloptions.c
+++ b/src/backend/access/common/reloptions.c
@@ -122,6 +122,15 @@ static relopt_bool boolRelOpts[] =
 		},
 		false
 	},
+	{
+		{
+			"publish_via_parent",
+			"Replicate this table's contents via its parent table when publishing with publish_via_partition_root",
+			RELOPT_KIND_HEAP,
+			AccessExclusiveLock
+		},
+		false
+	},
 	{
 		{
 			"fastupdate",
@@ -1877,6 +1886,8 @@ default_reloptions(Datum reloptions, bool validate, relopt_kind kind)
 		offsetof(StdRdOptions, autovacuum) + offsetof(AutoVacOpts, analyze_scale_factor)},
 		{"user_catalog_table", RELOPT_TYPE_BOOL,
 		offsetof(StdRdOptions, user_catalog_table)},
+		{"publish_via_parent", RELOPT_TYPE_BOOL,
+		offsetof(StdRdOptions, publish_via_parent)},
 		{"parallel_workers", RELOPT_TYPE_INT,
 		offsetof(StdRdOptions, parallel_workers)},
 		{"vacuum_index_cleanup", RELOPT_TYPE_ENUM,
diff --git a/src/backend/catalog/pg_inherits.c b/src/backend/catalog/pg_inherits.c
index da969bd2f9..cf90769457 100644
--- a/src/backend/catalog/pg_inherits.c
+++ b/src/backend/catalog/pg_inherits.c
@@ -24,14 +24,22 @@
 #include "access/table.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_inherits.h"
+#include "catalog/partition.h"
+#include "miscadmin.h"
 #include "parser/parse_type.h"
 #include "storage/lmgr.h"
+#include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
+#include "utils/fmgrprotos.h"
+#include "utils/lsyscache.h"
 #include "utils/memutils.h"
 #include "utils/snapmgr.h"
 #include "utils/syscache.h"
 
+static List * find_all_inheritors_internal(Oid parentrelId, LOCKMODE lockmode,
+										   List **numparents, bool logical_only);
+
 /*
  * Entry of a hash table used in find_all_inheritors. See below.
  */
@@ -59,14 +67,14 @@ List *
 find_inheritance_children(Oid parentrelId, LOCKMODE lockmode)
 {
 	return find_inheritance_children_extended(parentrelId, true, lockmode,
-											  NULL, NULL);
+											  NULL, NULL, false);
 }
 
 /*
  * find_inheritance_children_extended
  *
  * As find_inheritance_children, with more options regarding detached
- * partitions.
+ * partitions and logical roots.
  *
  * If a partition's pg_inherits row is marked "detach pending",
  * *detached_exist (if not null) is set true.
@@ -78,16 +86,21 @@ find_inheritance_children(Oid parentrelId, LOCKMODE lockmode)
  * whether the transaction that marked those partitions as detached appears
  * committed to the active snapshot.  In addition, *detached_xmin (if not null)
  * is set to the xmin of the row of the detached partition.
+ *
+ * If logical_only is true, only tables marked explicitly via publish_via_parent
+ * are included in the output list.
  */
 List *
 find_inheritance_children_extended(Oid parentrelId, bool omit_detached,
 								   LOCKMODE lockmode, bool *detached_exist,
-								   TransactionId *detached_xmin)
+								   TransactionId *detached_xmin,
+								   bool logical_only)
 {
 	List	   *list = NIL;
 	Relation	relation;
+	Oid			index;
 	SysScanDesc scan;
-	ScanKeyData key[1];
+	ScanKeyData key[2];
 	HeapTuple	inheritsTuple;
 	Oid			inhrelid;
 	Oid		   *oidarr;
@@ -110,14 +123,24 @@ find_inheritance_children_extended(Oid parentrelId, bool omit_detached,
 	numoids = 0;
 
 	relation = table_open(InheritsRelationId, AccessShareLock);
+	index = InheritsParentIndexId;
 
 	ScanKeyInit(&key[0],
 				Anum_pg_inherits_inhparent,
 				BTEqualStrategyNumber, F_OIDEQ,
 				ObjectIdGetDatum(parentrelId));
 
-	scan = systable_beginscan(relation, InheritsParentIndexId, true,
-							  NULL, 1, key);
+	if (logical_only)
+	{
+		index = InheritsParentSeqnoIndexId;
+		ScanKeyInit(&key[1],
+					Anum_pg_inherits_inhseqno,
+					BTEqualStrategyNumber, F_INT4EQ,
+					Int32GetDatum(0));
+	}
+
+	scan = systable_beginscan(relation, index, true,
+							  NULL, logical_only ? 2 : 1, key);
 
 	while ((inheritsTuple = systable_getnext(scan)) != NULL)
 	{
@@ -254,6 +277,20 @@ find_inheritance_children_extended(Oid parentrelId, bool omit_detached,
  */
 List *
 find_all_inheritors(Oid parentrelId, LOCKMODE lockmode, List **numparents)
+{
+	return find_all_inheritors_internal(parentrelId, lockmode, numparents,
+										false);
+}
+
+List *
+find_all_logical_inheritors(Oid parentrelId)
+{
+	return find_all_inheritors_internal(parentrelId, NoLock, NULL, true);
+}
+
+static List *
+find_all_inheritors_internal(Oid parentrelId, LOCKMODE lockmode,
+							 List **numparents, bool logical_only)
 {
 	/* hash table for O(1) rel_oid -> rel_numparents cell lookup */
 	HTAB	   *seen_rels;
@@ -290,7 +327,9 @@ find_all_inheritors(Oid parentrelId, LOCKMODE lockmode, List **numparents)
 		ListCell   *lc;
 
 		/* Get the direct children of this rel */
-		currentchildren = find_inheritance_children(currentrel, lockmode);
+		currentchildren =
+			find_inheritance_children_extended(currentrel, true, lockmode,
+											   NULL, NULL, logical_only);
 
 		/*
 		 * Add to the queue only those children not already seen. This avoids
@@ -655,3 +694,73 @@ PartitionHasPendingDetach(Oid partoid)
 	elog(ERROR, "relation %u is not a partition", partoid);
 	return false;				/* keep compiler quiet */
 }
+
+static Oid
+get_logical_parent_worker(Relation inhRel, Oid relid)
+{
+	SysScanDesc scan;
+	ScanKeyData key[2];
+	Oid			result = InvalidOid;
+	HeapTuple	tuple;
+
+	ScanKeyInit(&key[0],
+				Anum_pg_inherits_inhrelid,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(relid));
+	ScanKeyInit(&key[1],
+				Anum_pg_inherits_inhseqno,
+				BTEqualStrategyNumber, F_INT4EQ,
+				Int32GetDatum(0));
+
+	scan = systable_beginscan(inhRel, InheritsRelidSeqnoIndexId, true,
+							  NULL, 2, key);
+	tuple = systable_getnext(scan);
+	if (HeapTupleIsValid(tuple))
+	{
+		Form_pg_inherits form = (Form_pg_inherits) GETSTRUCT(tuple);
+		result = form->inhparent;
+	}
+
+	systable_endscan(scan);
+
+	return result;
+}
+
+static void
+get_logical_ancestors_worker(Relation inhRel, Oid relid, List **ancestors)
+{
+	Oid			parentOid;
+
+	/*
+	 * Recursion ends at the topmost level, ie., when there's no parent.
+	 */
+	parentOid = get_logical_parent_worker(inhRel, relid);
+	if (parentOid == InvalidOid)
+		return;
+
+	*ancestors = lappend_oid(*ancestors, parentOid);
+	get_logical_ancestors_worker(inhRel, parentOid, ancestors);
+}
+
+List *
+get_logical_ancestors(Oid relid, bool is_partition)
+{
+	List	   *result = NIL;
+	Relation	inhRel;
+
+	/* For partitions, this is identical to get_partition_ancestors(). */
+	if (is_partition)
+		return get_partition_ancestors(relid);
+
+	inhRel = table_open(InheritsRelationId, AccessShareLock);
+	get_logical_ancestors_worker(inhRel, relid, &result);
+	table_close(inhRel, AccessShareLock);
+
+	return result;
+}
+
+bool
+has_logical_parent(Relation inhRel, Oid relid)
+{
+	return (get_logical_parent_worker(inhRel, relid) != InvalidOid);
+}
diff --git a/src/backend/catalog/pg_publication.c b/src/backend/catalog/pg_publication.c
index df9abf09c3..20970a4eb0 100644
--- a/src/backend/catalog/pg_publication.c
+++ b/src/backend/catalog/pg_publication.c
@@ -51,6 +51,7 @@ typedef struct
 	Oid			relid;			/* OID of published table */
 	Oid			pubid;			/* OID of publication that publishes this
 								 * table. */
+	bool		viaroot;		/* does pubid use publish_via_partition_root? */
 } published_rel;
 
 static void publication_translate_columns(Relation targetrel, List *columns,
@@ -180,56 +181,46 @@ pg_relation_is_publishable(PG_FUNCTION_ARGS)
 }
 
 /*
- * Returns true if the ancestor is in the list of published relations.
- * Otherwise, returns false.
- */
-static bool
-is_ancestor_member_tableinfos(Oid ancestor, List *table_infos)
-{
-	ListCell   *lc;
-
-	foreach(lc, table_infos)
-	{
-		Oid			relid = ((published_rel *) lfirst(lc))->relid;
-
-		if (relid == ancestor)
-			return true;
-	}
-
-	return false;
-}
-
-/*
- * Filter out the partitions whose parent tables are also present in the list.
+ * Filter out the tables who are already being published via a logical root.
+ *
+ * If we only had partitioned tables to check, this would be much easier: it's
+ * impossible for a partition root to be in this list unless pubviaroot=true for
+ * that table, so we would just have to check for the existence of an ancestor,
+ * and if any were found then we'd filter the table.
+ *
+ * Logical roots add another wrinkle, because it's acceptable for a
+ * pubviaroot=false publication to contain a logical root. (Logical roots can
+ * contain data of their own, unlike partition roots.) It's also possible for a
+ * logical root to be included in a pubviaroot publication without any or all of
+ * its children. This is not possible for a partition root, where all of the
+ * leaves are implicitly included.
+ *
+ * So, to keep things relatively consistent, this now uses the same code that
+ * the output plugin uses to decide which relation OID to publish data under. If
+ * the publishing OID is not the same as the relation OID, we remove it from
+ * this list.
+ *
+ * XXX And that's expensive.
  */
 static void
-filter_partitions(List *table_infos)
+filter_published_descendants(List *table_infos, List *publications)
 {
 	ListCell   *lc;
 
 	foreach(lc, table_infos)
 	{
-		bool		skip = false;
-		List	   *ancestors = NIL;
-		ListCell   *lc2;
 		published_rel *table_info = (published_rel *) lfirst(lc);
+		Oid			publish_as_relid;
+		List	   *rel_publications;
 
-		if (get_rel_relispartition(table_info->relid))
-			ancestors = get_partition_ancestors(table_info->relid);
-
-		foreach(lc2, ancestors)
-		{
-			Oid			ancestor = lfirst_oid(lc2);
-
-			if (is_ancestor_member_tableinfos(ancestor, table_infos))
-			{
-				skip = true;
-				break;
-			}
-		}
+		rel_publications =
+			process_relation_publications(table_info->relid, publications, NULL,
+										  &publish_as_relid);
 
-		if (skip)
+		if (publish_as_relid != table_info->relid)
 			table_infos = foreach_delete_current(table_infos, lc);
+
+		list_free(rel_publications);
 	}
 }
 
@@ -805,11 +796,15 @@ List *
 GetAllTablesPublicationRelations(bool pubviaroot)
 {
 	Relation	classRel;
+	Relation	inhRel;
 	ScanKeyData key[1];
 	TableScanDesc scan;
 	HeapTuple	tuple;
 	List	   *result = NIL;
 
+	/* TODO: is there a required order to acquire these locks? */
+	if (pubviaroot)
+		inhRel = table_open(InheritsRelationId, AccessShareLock);
 	classRel = table_open(RelationRelationId, AccessShareLock);
 
 	ScanKeyInit(&key[0],
@@ -825,7 +820,8 @@ GetAllTablesPublicationRelations(bool pubviaroot)
 		Oid			relid = relForm->oid;
 
 		if (is_publishable_class(relid, relForm) &&
-			!(relForm->relispartition && pubviaroot))
+			!(relForm->relispartition && pubviaroot) &&
+			!(pubviaroot && has_logical_parent(inhRel, relid)))
 			result = lappend_oid(result, relid);
 	}
 
@@ -854,6 +850,9 @@ GetAllTablesPublicationRelations(bool pubviaroot)
 	}
 
 	table_close(classRel, AccessShareLock);
+	if (pubviaroot)
+		table_close(inhRel, AccessShareLock);
+
 	return result;
 }
 
@@ -1070,6 +1069,7 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 		int			nelems,
 					i;
 		bool		viaroot = false;
+		List	   *publications = NIL;
 
 		/* create a function context for cross-call persistence */
 		funcctx = SRF_FIRSTCALL_INIT();
@@ -1093,6 +1093,7 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 			ListCell   *lc;
 
 			pub_elem = GetPublicationByName(TextDatumGetCString(elems[i]), false);
+			publications = lappend(publications, pub_elem);
 
 			/*
 			 * Publications support partitioned tables. If
@@ -1132,6 +1133,7 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 
 				table_info->relid = lfirst_oid(lc);
 				table_info->pubid = pub_elem->oid;
+				table_info->viaroot = pub_elem->pubviaroot;
 				table_infos = lappend(table_infos, table_info);
 			}
 
@@ -1141,15 +1143,14 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 		}
 
 		/*
-		 * If the publication publishes partition changes via their respective
-		 * root partitioned tables, we must exclude partitions in favor of
-		 * including the root partitioned tables. Otherwise, the function
-		 * could return both the child and parent tables which could cause
-		 * data of the child table to be double-published on the subscriber
-		 * side.
+		 * If the publication publishes table changes via their respective
+		 * logical root tables, we must exclude partitions in favor of
+		 * including the root tables. Otherwise, the function could return
+		 * both the child and parent tables which could cause data of the
+		 * child table to be double-published on the subscriber side.
 		 */
 		if (viaroot)
-			filter_partitions(table_infos);
+			filter_published_descendants(table_infos, publications);
 
 		/* Construct a tuple descriptor for the result rows. */
 		tupdesc = CreateTemplateTupleDesc(NUM_PUBLICATION_TABLES_ELEM);
@@ -1165,6 +1166,8 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 		funcctx->tuple_desc = BlessTupleDesc(tupdesc);
 		funcctx->user_fctx = (void *) table_infos;
 
+		list_free(publications);
+
 		MemoryContextSwitchTo(oldcontext);
 	}
 
@@ -1260,6 +1263,146 @@ pg_get_publication_tables(PG_FUNCTION_ARGS)
 	SRF_RETURN_DONE(funcctx);
 }
 
+/*
+ * Returns a list of tables that should be copied during initial sync for the
+ * given root table and publications.
+ *
+ * The only time this list consists of anything more than the table itself is
+ * when a publication's publish_via_partition_root is set to true and the table
+ * has inherited child tables in the publications that have been marked with
+ * publish_via_parent.
+ */
+Datum
+pg_get_publication_rels_to_sync(PG_FUNCTION_ARGS)
+{
+	FuncCallContext *funcctx;
+	List	   *tables = NIL;
+
+	/* stuff done only on the first call of the function */
+	if (SRF_IS_FIRSTCALL())
+	{
+		Oid			rootid = PG_GETARG_OID(0);
+		MemoryContext oldcontext;
+		List	   *publications = NIL;
+		List	   *inheritors;
+		List	   *candidates = NIL;
+		ListCell   *lc;
+
+		/* create a function context for cross-call persistence */
+		funcctx = SRF_FIRSTCALL_INIT();
+
+		/* switch to memory context appropriate for multiple function calls */
+		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
+
+		/* Construct the list of publications to consider. */
+		{
+			ArrayType  *arr;
+			Datum	   *elems;
+			int			nelems,
+						i;
+
+			/*
+			 * Deconstruct the parameter into elements where each element is a
+			 * publication name.
+			 */
+			arr = PG_GETARG_ARRAYTYPE_P(1);
+			deconstruct_array(arr, TEXTOID, -1, false, TYPALIGN_INT,
+							  &elems, NULL, &nelems);
+
+			for (i = 0; i < nelems; i++)
+			{
+				char	   *pubname = TextDatumGetCString(elems[i]);
+				Publication *pub = GetPublicationByName(pubname, false);
+
+				publications = lappend(publications, pub);
+			}
+		}
+
+		/* TODO: do the tables in this list need to be locked? */
+		inheritors = find_all_logical_inheritors(rootid);
+
+		/*
+		 * For a logical descendant to be in the list to sync, both it and the
+		 * root table need to be part of a pubviaroot publication. (But they
+		 * don't have to be part of the *same* pubviaroot publication. This is
+		 * somewhat counterintuitive, but it prevents a descendant from being
+		 * double-published through two separate roots.)
+		 */
+		foreach(lc, publications)
+		{
+			Publication *publication = lfirst(lc);
+			List	   *relids;
+			List	   *schemarelids;
+			List	   *published;
+			ListCell   *cell;
+
+			if (!publication->pubviaroot)
+				continue;
+
+			if (publication->alltables)
+			{
+				/* Easy case: every inheritor is published through the root. */
+				candidates = inheritors;
+				break;
+			}
+
+			/*
+			 * Figure out which of the inheritors is included in this
+			 * publication. They get added to our candidates list.
+			 */
+			relids = GetPublicationRelations(publication->oid,
+											 publication->pubviaroot ?
+											 PUBLICATION_PART_ROOT :
+											 PUBLICATION_PART_ALL);
+			schemarelids = GetAllSchemaPublicationRelations(publication->oid,
+															publication->pubviaroot ?
+															PUBLICATION_PART_ROOT :
+															PUBLICATION_PART_LEAF);
+			published = list_concat_unique_oid(relids, schemarelids);
+
+			foreach(cell, inheritors)
+			{
+				Oid		current = lfirst_oid(cell);
+
+				if (list_member_oid(published, current))
+					candidates = list_append_unique_oid(candidates, current);
+			}
+		}
+
+		/*
+		 * Finally, if we didn't find the root table itself in a pubviaroot
+		 * publication, none of the logical descendants will be published
+		 * through it.
+		 *
+		 * XXX: this assumes that the root table was part of at least one of the
+		 * publications in the list. It doesn't make any sense to call this
+		 * function otherwise, but should we ERROR out if that assumption is
+		 * violated?
+		 */
+		if (list_member_oid(candidates, rootid))
+			tables = candidates;
+		else
+			tables = list_make1_oid(rootid);
+
+		funcctx->user_fctx = (void *) tables;
+
+		MemoryContextSwitchTo(oldcontext);
+	}
+
+	/* stuff done on every call of the function */
+	funcctx = SRF_PERCALL_SETUP();
+	tables = (List *) funcctx->user_fctx;
+
+	if (funcctx->call_cntr < list_length(tables))
+	{
+		Oid			relid = list_nth_oid(tables, funcctx->call_cntr);
+
+		SRF_RETURN_NEXT(funcctx, ObjectIdGetDatum(relid));
+	}
+
+	SRF_RETURN_DONE(funcctx);
+}
+
 List *
 process_relation_publications(Oid relid, const List *publications,
 							  PublicationActions *pubactions,
@@ -1277,10 +1420,31 @@ process_relation_publications(Oid relid, const List *publications,
 	ListCell   *lc;
 	int			publish_ancestor_level = 0;
 	bool		am_partition = get_rel_relispartition(relid);
+	bool		am_pubviaroot = false;
 	char		relkind = get_rel_relkind(relid);
 	List	   *rel_publications = NIL;
+	Oid			publish_candidate = relid;
+
+	/*
+	 * For publish_via_parent handling later, it's useful to know whether the
+	 * current table is explicitly part of a publish_via_partition_root
+	 * publication up front.
+	 */
+	foreach(lc, publications)
+	{
+		Publication *pub = lfirst(lc);
 
-	*publish_as_relid = relid;
+		if (!pub->pubviaroot)
+			continue;
+
+		if (pub->alltables ||
+			list_member_oid(pubids, pub->oid) ||
+			list_member_oid(schemaPubids, pub->oid))
+		{
+			am_pubviaroot = true;
+			break;
+		}
+	}
 
 	foreach(lc, publications)
 	{
@@ -1296,55 +1460,57 @@ process_relation_publications(Oid relid, const List *publications,
 		int			ancestor_level = 0;
 
 		/*
-		 * If this is a FOR ALL TABLES publication, pick the partition
+		 * If this is a FOR ALL TABLES publication, pick the logical
 		 * root and set the ancestor level accordingly.
 		 */
 		if (pub->alltables)
 		{
 			publish = true;
-			if (pub->pubviaroot && am_partition)
+			if (pub->pubviaroot)
 			{
-				List	   *ancestors = get_partition_ancestors(relid);
+				List	   *ancestors;
 
-				pub_relid = llast_oid(ancestors);
-				ancestor_level = list_length(ancestors);
+				ancestors = get_logical_ancestors(relid, am_partition);
+				if (ancestors != NIL)
+				{
+					pub_relid = llast_oid(ancestors);
+					ancestor_level = list_length(ancestors);
+				}
 			}
 		}
-
-		if (!publish)
+		else
 		{
 			bool		ancestor_published = false;
+			Oid			ancestor;
+			int			level;
+			List	   *ancestors;
 
 			/*
-			 * For a partition, check if any of the ancestors are
-			 * published.  If so, note down the topmost ancestor that is
+			 * Check if any of the logical ancestors (that is, partition
+			 * parents or tables marked with publish_via_parent) are
+			 * published. If so, note down the topmost ancestor that is
 			 * published via this publication, which will be used as the
-			 * relation via which to publish the partition's changes.
+			 * relation via which to publish this table's changes.
 			 */
-			if (am_partition)
-			{
-				Oid			ancestor;
-				int			level;
-				List	   *ancestors = get_partition_ancestors(relid);
+			ancestors = get_logical_ancestors(relid, am_partition);
+			ancestor = GetTopMostAncestorInPublication(pub->oid,
+													   ancestors,
+													   &level);
 
-				ancestor = GetTopMostAncestorInPublication(pub->oid,
-														   ancestors,
-														   &level);
-
-				if (ancestor != InvalidOid)
+			if (ancestor != InvalidOid)
+			{
+				ancestor_published = true;
+				if (pub->pubviaroot)
 				{
-					ancestor_published = true;
-					if (pub->pubviaroot)
-					{
-						pub_relid = ancestor;
-						ancestor_level = level;
-					}
+					pub_relid = ancestor;
+					ancestor_level = level;
 				}
 			}
 
 			if (list_member_oid(pubids, pub->oid) ||
 				list_member_oid(schemaPubids, pub->oid) ||
-				ancestor_published)
+				(ancestor_published && am_partition) ||
+				(ancestor_published && am_pubviaroot && pub->pubviaroot))
 				publish = true;
 		}
 
@@ -1359,10 +1525,13 @@ process_relation_publications(Oid relid, const List *publications,
 		if (publish &&
 			(relkind != RELKIND_PARTITIONED_TABLE || pub->pubviaroot))
 		{
-			pubactions->pubinsert |= pub->pubactions.pubinsert;
-			pubactions->pubupdate |= pub->pubactions.pubupdate;
-			pubactions->pubdelete |= pub->pubactions.pubdelete;
-			pubactions->pubtruncate |= pub->pubactions.pubtruncate;
+			if (pubactions)
+			{
+				pubactions->pubinsert |= pub->pubactions.pubinsert;
+				pubactions->pubupdate |= pub->pubactions.pubupdate;
+				pubactions->pubdelete |= pub->pubactions.pubdelete;
+				pubactions->pubtruncate |= pub->pubactions.pubtruncate;
+			}
 
 			/*
 			 * We want to publish the changes as the top-most ancestor
@@ -1381,7 +1550,7 @@ process_relation_publications(Oid relid, const List *publications,
 			 */
 			if (publish_ancestor_level < ancestor_level)
 			{
-				*publish_as_relid = pub_relid;
+				publish_candidate = pub_relid;
 				publish_ancestor_level = ancestor_level;
 
 				/* reset the publication list for this relation */
@@ -1390,7 +1559,7 @@ process_relation_publications(Oid relid, const List *publications,
 			else
 			{
 				/* Same ancestor level, has to be the same OID. */
-				Assert(*publish_as_relid == pub_relid);
+				Assert(publish_candidate == pub_relid);
 			}
 
 			/* Track publications for this ancestor. */
@@ -1401,6 +1570,9 @@ process_relation_publications(Oid relid, const List *publications,
 	list_free(pubids);
 	list_free(schemaPubids);
 
+	if (publish_as_relid)
+		*publish_as_relid = publish_candidate;
+
 	return rel_publications;
 }
 
@@ -1434,7 +1606,7 @@ pg_get_relation_publishing_info(PG_FUNCTION_ARGS)
 
 	{
 		List	   *rel_publications;
-		PublicationActions pubactions;
+		PublicationActions pubactions = {0};
 		Oid			publish_as_relid;
 		ListCell   *lc;
 		Datum	   *puboids;
diff --git a/src/backend/commands/publicationcmds.c b/src/backend/commands/publicationcmds.c
index f4ba572697..a7f66d2982 100644
--- a/src/backend/commands/publicationcmds.c
+++ b/src/backend/commands/publicationcmds.c
@@ -238,6 +238,8 @@ contain_invalid_rfcolumn_walker(Node *node, rf_context *context)
 		 * parent table, but the bitmap contains the replica identity
 		 * information of the child table. So, get the column number of the
 		 * child table as parent and child column order could be different.
+		 *
+		 * TODO: is this applicable to publish_via_parent?
 		 */
 		if (context->pubviaroot)
 		{
@@ -286,6 +288,8 @@ pub_rf_contains_invalid_column(Oid pubid, Relation relation, List *ancestors,
 	 *
 	 * Note that even though the row filter used is for an ancestor, the
 	 * REPLICA IDENTITY used will be for the actual child table.
+	 *
+	 * TODO: is this applicable to publish_via_parent?
 	 */
 	if (pubviaroot && relation->rd_rel->relispartition)
 	{
@@ -336,6 +340,8 @@ pub_rf_contains_invalid_column(Oid pubid, Relation relation, List *ancestors,
  * the column list.
  *
  * Returns true if any replica identity column is not covered by column list.
+ *
+ * TODO: publish_via_parent?
  */
 bool
 pub_collist_contains_invalid_column(Oid pubid, Relation relation, List *ancestors,
@@ -628,6 +634,9 @@ TransformPubWhereClauses(List *tables, const char *queryString,
 		 * If the publication doesn't publish changes via the root partitioned
 		 * table, the partition's row filter will be used. So disallow using
 		 * WHERE clause on partitioned table in this case.
+		 *
+		 * TODO: decide how this interacts with publish_via_parent. Probably not
+		 * in the same way, since logical roots can contain data of their own.
 		 */
 		if (!pubviaroot &&
 			pri->relation->rd_rel->relkind == RELKIND_PARTITIONED_TABLE)
@@ -715,6 +724,9 @@ CheckPubRelationColumnList(char *pubname, List *tables,
 		 * If the publication doesn't publish changes via the root partitioned
 		 * table, the partition's column list will be used. So disallow using
 		 * a column list on the partitioned table in this case.
+		 *
+		 * TODO: decide if this interacts with publish_via_parent. Probably not
+		 * (see above).
 		 */
 		if (!pubviaroot &&
 			pri->relation->rd_rel->relkind == RELKIND_PARTITIONED_TABLE)
diff --git a/src/backend/commands/tablecmds.c b/src/backend/commands/tablecmds.c
index d097da3c78..43ae131021 100644
--- a/src/backend/commands/tablecmds.c
+++ b/src/backend/commands/tablecmds.c
@@ -357,7 +357,8 @@ static bool MergeCheckConstraint(List *constraints, char *name, Node *expr);
 static void MergeAttributesIntoExisting(Relation child_rel, Relation parent_rel);
 static void MergeConstraintsIntoExisting(Relation child_rel, Relation parent_rel);
 static void StoreCatalogInheritance(Oid relationId, List *supers,
-									bool child_is_partition);
+									bool child_is_partition,
+									bool publish_via_parent);
 static void StoreCatalogInheritance1(Oid relationId, Oid parentOid,
 									 int32 seqNumber, Relation inhRelation,
 									 bool child_is_partition);
@@ -586,6 +587,7 @@ static void ATPrepSetTableSpace(AlteredTableInfo *tab, Relation rel,
 								const char *tablespacename, LOCKMODE lockmode);
 static void ATExecSetTableSpace(Oid tableOid, Oid newTableSpace, LOCKMODE lockmode);
 static void ATExecSetTableSpaceNoStorage(Relation rel, Oid newTableSpace);
+static void change_publish_via_parent(Relation rel, bool set);
 static void ATExecSetRelOptions(Relation rel, List *defList,
 								AlterTableType operation,
 								LOCKMODE lockmode);
@@ -1123,7 +1125,8 @@ DefineRelation(CreateStmt *stmt, char relkind, Oid ownerId,
 	}
 
 	/* Store inheritance information for new rel. */
-	StoreCatalogInheritance(relationId, inheritOids, stmt->partbound != NULL);
+	StoreCatalogInheritance(relationId, inheritOids, stmt->partbound != NULL,
+							RelationIsPublishedViaParent(rel));
 
 	/*
 	 * Process the partitioning specification (if any) and store the partition
@@ -3325,7 +3328,7 @@ MergeCheckConstraint(List *constraints, char *name, Node *expr)
  */
 static void
 StoreCatalogInheritance(Oid relationId, List *supers,
-						bool child_is_partition)
+						bool child_is_partition, bool publish_via_parent)
 {
 	Relation	relation;
 	int32		seqNumber;
@@ -3336,6 +3339,22 @@ StoreCatalogInheritance(Oid relationId, List *supers,
 	 */
 	Assert(OidIsValid(relationId));
 
+	/*
+	 * publish_via_parent requires exactly one parent. Try to keep this
+	 * messaging consistent with ALTER TABLE/change_publish_via_parent().
+	 *
+	 * Earlier, in MergeAttributes, we already checked that we had ownership of
+	 * the parent table, so we don't need to check that again the way that
+	 * change_publish_via_parent() does.
+	 */
+	if (publish_via_parent && list_length(supers) != 1)
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg((list_length(supers) == 0)
+						? "table \"%s\" does not inherit from any tables"
+						: "table \"%s\" inherits from multiple tables",
+						get_rel_name(relationId))));
+
 	if (supers == NIL)
 		return;
 
@@ -3350,7 +3369,12 @@ StoreCatalogInheritance(Oid relationId, List *supers,
 	 */
 	relation = table_open(InheritsRelationId, RowExclusiveLock);
 
-	seqNumber = 1;
+	/*
+	 * Normally inhseqno starts at one, but for publish_via_parent, it's given a
+	 * sentinel value of zero. (In that case, we'll only go through this loop
+	 * once, as the list_length() check above enforces.)
+	 */
+	seqNumber = publish_via_parent ? 0 : 1;
 	foreach(entry, supers)
 	{
 		Oid			parentOid = lfirst_oid(entry);
@@ -14829,6 +14853,88 @@ ATPrepSetTableSpace(AlteredTableInfo *tab, Relation rel, const char *tablespacen
 	tab->newTableSpace = tablespaceId;
 }
 
+static void
+change_publish_via_parent(Relation rel, bool set)
+{
+	Relation	inhRel;
+	ScanKeyData key;
+	SysScanDesc scan;
+	Oid			parentOid;
+	Relation	parentRel;
+	HeapTuple	tuple, copyTuple;
+	Form_pg_inherits form;
+
+	/* Open pg_inherits with RowExclusiveLock so that we can update it. */
+	inhRel = table_open(InheritsRelationId, RowExclusiveLock);
+
+	/*
+	 * We have to make sure that an inheritance relationship already exists, and
+	 * that there is only one candidate parent for this table.
+	 *
+	 * TODO: do we have to lock the tables themselves to avoid races?
+	 */
+	ScanKeyInit(&key,
+				Anum_pg_inherits_inhrelid,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(RelationGetRelid(rel)));
+
+	scan = systable_beginscan(inhRel, InheritsRelidSeqnoIndexId, true,
+							  NULL, 1, &key);
+	tuple = systable_getnext(scan);
+
+	if (!HeapTupleIsValid(tuple))
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("table \"%s\" does not inherit from any tables",
+						RelationGetRelationName(rel))));
+
+	form = (Form_pg_inherits) GETSTRUCT(tuple);
+	parentOid = form->inhparent;
+	copyTuple = heap_copytuple(tuple);
+
+	if (HeapTupleIsValid(systable_getnext(scan)))
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("table \"%s\" inherits from multiple tables",
+						RelationGetRelationName(rel))));
+
+	systable_endscan(scan);
+
+	if (set)
+	{
+		/*
+		 * Open our parent table and check permissions. Like ATTACH PARTITION,
+		 * we require ownership of the parent table in addition to the child
+		 * (which has already been checked in ATPrepCmd).
+		 *
+		 * TODO: what lock strength is needed here? Is nesting it inside the
+		 * inhRel lock a risk?
+		 */
+		parentRel = table_open(parentOid, AccessExclusiveLock);
+		ATSimplePermissions(AT_SetRelOptions, parentRel, ATT_TABLE);
+		table_close(parentRel, NoLock); /* TODO: do we need to hang onto the lock? */
+	}
+	else
+	{
+		/*
+		 * Ownership of the parent isn't needed for unsetting the flag, just
+		 * like we wouldn't check ownership during NO INHERIT.
+		 */
+	}
+
+	/*
+	 * Mark the inheritance as a logical root by setting inhseqno to zero.
+	 * Unmark it by setting inhseqno to one (since there's only one parent).
+	 */
+	form = (Form_pg_inherits) GETSTRUCT(copyTuple);
+	form->inhseqno = set ? 0 : 1;
+
+	CatalogTupleUpdate(inhRel, &copyTuple->t_self, copyTuple);
+
+	heap_freetuple(copyTuple);
+	table_close(inhRel, RowExclusiveLock);
+}
+
 /*
  * Set, reset, or replace reloptions.
  */
@@ -14940,6 +15046,39 @@ ATExecSetRelOptions(Relation rel, List *defList, AlterTableType operation,
 		}
 	}
 
+	/* Extra work for publish_via_parent */
+	if (rel->rd_rel->relkind == RELKIND_RELATION)
+	{
+		List	   *heap_options = untransformRelOptions(newOptions);
+		ListCell   *cell;
+		bool		curSetting = RelationIsPublishedViaParent(rel);
+		DefElem    *pubElem = NULL;
+
+		foreach (cell, heap_options)
+		{
+			DefElem    *defel = lfirst_node(DefElem, cell);
+
+			if (strcmp(defel->defname, "publish_via_parent") == 0)
+			{
+				pubElem = defel;
+				break;
+			}
+		}
+
+		if (pubElem)
+		{
+			bool		newSetting = defGetBoolean(pubElem);
+
+			if (newSetting != curSetting)
+				change_publish_via_parent(rel, newSetting);
+		}
+		else if (curSetting)
+		{
+			/* publish_via_parent was RESET */
+			change_publish_via_parent(rel, false);
+		}
+	}
+
 	/*
 	 * All we need do here is update the pg_class row; the new options will be
 	 * propagated into relcaches during post-commit cache inval.
@@ -15538,6 +15677,13 @@ ATExecAddInherit(Relation child_rel, RangeVar *parent, LOCKMODE lockmode)
 						trigger_name, RelationGetRelationName(child_rel)),
 				 errdetail("ROW triggers with transition tables are not supported in inheritance hierarchies.")));
 
+	/* publish_via_parent tables are not allowed to add additional parents. */
+	if (RelationIsPublishedViaParent(child_rel))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("relation option \"%s\" prevents table \"%s\" from changing inheritance",
+						"publish_via_parent", RelationGetRelationName(child_rel))));
+
 	/* OK to create inheritance */
 	CreateInheritance(child_rel, parent_rel);
 
@@ -16001,6 +16147,13 @@ ATExecDropInherit(Relation rel, RangeVar *parent, LOCKMODE lockmode)
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
 				 errmsg("cannot change inheritance of a partition")));
 
+	/* publish_via_parent tables are not allowed to remove their parent. */
+	if (RelationIsPublishedViaParent(rel))
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("relation option \"%s\" prevents table \"%s\" from changing inheritance",
+						"publish_via_parent", RelationGetRelationName(rel))));
+
 	/*
 	 * AccessShareLock on the parent is probably enough, seeing that DROP
 	 * TABLE doesn't lock parent tables at all.  We need some lock since we'll
diff --git a/src/backend/partitioning/partdesc.c b/src/backend/partitioning/partdesc.c
index 65f3d5a5e6..0ed7cbbbc5 100644
--- a/src/backend/partitioning/partdesc.c
+++ b/src/backend/partitioning/partdesc.c
@@ -162,7 +162,8 @@ RelationBuildPartitionDesc(Relation rel, bool omit_detached)
 	inhoids = find_inheritance_children_extended(RelationGetRelid(rel),
 												 omit_detached, NoLock,
 												 &detached_exist,
-												 &detached_xmin);
+												 &detached_xmin,
+												 false);
 
 	nparts = list_length(inhoids);
 
diff --git a/src/backend/replication/logical/tablesync.c b/src/backend/replication/logical/tablesync.c
index e2cee92cf2..1e6dc70e32 100644
--- a/src/backend/replication/logical/tablesync.c
+++ b/src/backend/replication/logical/tablesync.c
@@ -771,11 +771,12 @@ copy_read_data(void *outbuf, int minread, int maxread)
 /*
  * Get information about remote relation in similar fashion the RELATION
  * message provides during replication. This function also returns the relation
- * qualifications to be used in the COPY command.
+ * qualifications to be used in the COPY command, and the list of tables to COPY
+ * (which for most tables will contain only one entry).
  */
 static void
 fetch_remote_table_info(char *nspname, char *relname,
-						LogicalRepRelation *lrel, List **qual)
+						LogicalRepRelation *lrel, List **qual, List **to_copy)
 {
 	WalRcvExecResult *res;
 	StringInfoData cmd;
@@ -1086,6 +1087,87 @@ fetch_remote_table_info(char *nspname, char *relname,
 		walrcv_clear_result(res);
 	}
 
+	/*
+	 * See if there are any other tables to be copied besides the original. This
+	 * happens when a descendant in the inheritance relationship is marked with
+	 * publish_via_parent and is part of a publication with
+	 * publish_via_partition_root = true.
+	 *
+	 * FIXME: if two publications have conflicting settings for
+	 * publish_via_partition_root, this behaves strangely. It looks like that's
+	 * true for regular partitions too...
+	 */
+	if (walrcv_server_version(LogRepWorkerWalRcvConn) >= 160000)
+	{
+		Oid			descRow[] = {TEXTOID, TEXTOID};
+		StringInfoData pub_names;
+
+		/* Build the pubname list. */
+		initStringInfo(&pub_names);
+		foreach(lc, MySubscription->publications)
+		{
+			char	   *pubname = strVal(lfirst(lc));
+
+			if (foreach_current_index(lc) > 0)
+				appendStringInfoString(&pub_names, ", ");
+
+			appendStringInfoString(&pub_names, quote_literal_cstr(pubname));
+		}
+
+		/*
+		 * Ask the server what it wants us to sync.
+		 */
+		resetStringInfo(&cmd);
+		appendStringInfo(&cmd,
+						 "SELECT DISTINCT n.nspname, c.relname"
+						 "  FROM pg_catalog.pg_publication p,"
+						 "       pg_catalog.pg_class c"
+						 "  JOIN pg_catalog.pg_namespace n"
+						 "    ON (c.relnamespace = n.oid)"
+						 "  JOIN LATERAL pg_catalog.pg_get_publication_rels_to_sync(%u::regclass, p.pubname) relid"
+						 "    ON (c.oid = relid)"
+						 " WHERE p.pubname IN ( %s )",
+						 lrel->remoteid, pub_names.data);
+
+		res = walrcv_exec(LogRepWorkerWalRcvConn, cmd.data,
+						  lengthof(descRow), descRow);
+
+		if (res->status != WALRCV_OK_TUPLES)
+			ereport(ERROR,
+					(errmsg("could not fetch logical descendants for table \"%s.%s\" from publisher: %s",
+							nspname, relname, res->err)));
+
+		slot = MakeSingleTupleTableSlot(res->tupledesc, &TTSOpsMinimalTuple);
+		while (tuplestore_gettupleslot(res->tuplestore, true, false, slot))
+		{
+			char	   *desc_nspname;
+			char	   *desc_relname;
+			char	   *quoted;
+
+			desc_nspname = TextDatumGetCString(slot_getattr(slot, 1, &isnull));
+			Assert(!isnull);
+			desc_relname = TextDatumGetCString(slot_getattr(slot, 2, &isnull));
+			Assert(!isnull);
+
+			quoted = quote_qualified_identifier(desc_nspname, desc_relname);
+			*to_copy = lappend(*to_copy, quoted);
+
+			ExecClearTuple(slot);
+		}
+
+		ExecDropSingleTupleTableSlot(slot);
+		walrcv_clear_result(res);
+
+		pfree(pub_names.data);
+	}
+	else
+	{
+		/* For older servers, we only COPY the table itself. */
+		char   *quoted = quote_qualified_identifier(lrel->nspname,
+													lrel->relname);
+		*to_copy = lappend(*to_copy, quoted);
+	}
+
 	pfree(cmd.data);
 }
 
@@ -1100,6 +1182,8 @@ copy_table(Relation rel)
 	LogicalRepRelMapEntry *relmapentry;
 	LogicalRepRelation lrel;
 	List	   *qual = NIL;
+	List	   *to_copy = NIL;
+	ListCell   *cur;
 	WalRcvExecResult *res;
 	StringInfoData cmd;
 	CopyFromState cstate;
@@ -1109,7 +1193,8 @@ copy_table(Relation rel)
 
 	/* Get the publisher relation info. */
 	fetch_remote_table_info(get_namespace_name(RelationGetNamespace(rel)),
-							RelationGetRelationName(rel), &lrel, &qual);
+							RelationGetRelationName(rel), &lrel, &qual,
+							&to_copy);
 
 	/* Put the relation into relmap. */
 	logicalrep_relmap_update(&lrel);
@@ -1118,105 +1203,109 @@ copy_table(Relation rel)
 	relmapentry = logicalrep_rel_open(lrel.remoteid, NoLock);
 	Assert(rel == relmapentry->localrel);
 
-	/* Start copy on the publisher. */
-	initStringInfo(&cmd);
-
-	/* Regular table with no row filter */
-	if (lrel.relkind == RELKIND_RELATION && qual == NIL)
+	foreach(cur, to_copy)
 	{
-		appendStringInfo(&cmd, "COPY %s (",
-						 quote_qualified_identifier(lrel.nspname, lrel.relname));
-
-		/*
-		 * XXX Do we need to list the columns in all cases? Maybe we're
-		 * replicating all columns?
-		 */
-		for (int i = 0; i < lrel.natts; i++)
-		{
-			if (i > 0)
-				appendStringInfoString(&cmd, ", ");
+		char	   *quoted_name = lfirst(cur);
 
-			appendStringInfoString(&cmd, quote_identifier(lrel.attnames[i]));
-		}
+		/* Start copy on the publisher. */
+		initStringInfo(&cmd);
 
-		appendStringInfoString(&cmd, ") TO STDOUT");
-	}
-	else
-	{
-		/*
-		 * For non-tables and tables with row filters, we need to do COPY
-		 * (SELECT ...), but we can't just do SELECT * because we need to not
-		 * copy generated columns. For tables with any row filters, build a
-		 * SELECT query with OR'ed row filters for COPY.
-		 */
-		appendStringInfoString(&cmd, "COPY (SELECT ");
-		for (int i = 0; i < lrel.natts; i++)
+		/* Regular table with no row filter */
+		if (lrel.relkind == RELKIND_RELATION && qual == NIL)
 		{
-			appendStringInfoString(&cmd, quote_identifier(lrel.attnames[i]));
-			if (i < lrel.natts - 1)
-				appendStringInfoString(&cmd, ", ");
-		}
+			appendStringInfo(&cmd, "COPY %s (", quoted_name);
 
-		appendStringInfoString(&cmd, " FROM ");
+			/*
+			 * XXX Do we need to list the columns in all cases? Maybe we're
+			 * replicating all columns?
+			 */
+			for (int i = 0; i < lrel.natts; i++)
+			{
+				if (i > 0)
+					appendStringInfoString(&cmd, ", ");
 
-		/*
-		 * For regular tables, make sure we don't copy data from a child that
-		 * inherits the named table as those will be copied separately.
-		 */
-		if (lrel.relkind == RELKIND_RELATION)
-			appendStringInfoString(&cmd, "ONLY ");
+				appendStringInfoString(&cmd, quote_identifier(lrel.attnames[i]));
+			}
 
-		appendStringInfoString(&cmd, quote_qualified_identifier(lrel.nspname, lrel.relname));
-		/* list of OR'ed filters */
-		if (qual != NIL)
+			appendStringInfoString(&cmd, ") TO STDOUT");
+		}
+		else
 		{
-			ListCell   *lc;
-			char	   *q = strVal(linitial(qual));
+			/*
+			 * For non-tables and tables with row filters, we need to do COPY
+			 * (SELECT ...), but we can't just do SELECT * because we need to not
+			 * copy generated columns. For tables with any row filters, build a
+			 * SELECT query with OR'ed row filters for COPY.
+			 */
+			appendStringInfoString(&cmd, "COPY (SELECT ");
+			for (int i = 0; i < lrel.natts; i++)
+			{
+				appendStringInfoString(&cmd, quote_identifier(lrel.attnames[i]));
+				if (i < lrel.natts - 1)
+					appendStringInfoString(&cmd, ", ");
+			}
+
+			appendStringInfoString(&cmd, " FROM ");
+
+			/*
+			 * For regular tables, make sure we don't copy data from a child that
+			 * inherits the named table as those will be copied separately.
+			 */
+			if (lrel.relkind == RELKIND_RELATION)
+				appendStringInfoString(&cmd, "ONLY ");
 
-			appendStringInfo(&cmd, " WHERE %s", q);
-			for_each_from(lc, qual, 1)
+			appendStringInfoString(&cmd, quoted_name);
+			/* list of OR'ed filters */
+			if (qual != NIL)
 			{
-				q = strVal(lfirst(lc));
-				appendStringInfo(&cmd, " OR %s", q);
+				ListCell   *lc;
+				char	   *q = strVal(linitial(qual));
+
+				appendStringInfo(&cmd, " WHERE %s", q);
+				for_each_from(lc, qual, 1)
+				{
+					q = strVal(lfirst(lc));
+					appendStringInfo(&cmd, " OR %s", q);
+				}
+				list_free_deep(qual);
 			}
-			list_free_deep(qual);
-		}
 
-		appendStringInfoString(&cmd, ") TO STDOUT");
-	}
+			appendStringInfoString(&cmd, ") TO STDOUT");
+		}
 
-	/*
-	 * Prior to v16, initial table synchronization will use text format even
-	 * if the binary option is enabled for a subscription.
-	 */
-	if (walrcv_server_version(LogRepWorkerWalRcvConn) >= 160000 &&
-		MySubscription->binary)
-	{
-		appendStringInfoString(&cmd, " WITH (FORMAT binary)");
-		options = list_make1(makeDefElem("format",
-										 (Node *) makeString("binary"), -1));
-	}
+		/*
+		 * Prior to v16, initial table synchronization will use text format even
+		 * if the binary option is enabled for a subscription.
+		 */
+		if (walrcv_server_version(LogRepWorkerWalRcvConn) >= 160000 &&
+			MySubscription->binary)
+		{
+			appendStringInfoString(&cmd, " WITH (FORMAT binary)");
+			options = list_make1(makeDefElem("format",
+											 (Node *) makeString("binary"), -1));
+		}
 
-	res = walrcv_exec(LogRepWorkerWalRcvConn, cmd.data, 0, NULL);
-	pfree(cmd.data);
-	if (res->status != WALRCV_OK_COPY_OUT)
-		ereport(ERROR,
-				(errcode(ERRCODE_CONNECTION_FAILURE),
-				 errmsg("could not start initial contents copy for table \"%s.%s\": %s",
-						lrel.nspname, lrel.relname, res->err)));
-	walrcv_clear_result(res);
+		res = walrcv_exec(LogRepWorkerWalRcvConn, cmd.data, 0, NULL);
+		pfree(cmd.data);
+		if (res->status != WALRCV_OK_COPY_OUT)
+			ereport(ERROR,
+					(errcode(ERRCODE_CONNECTION_FAILURE),
+					 errmsg("could not start initial contents copy for table \"%s.%s\" from remote %s: %s",
+							lrel.nspname, lrel.relname, quoted_name, res->err)));
+		walrcv_clear_result(res);
 
-	copybuf = makeStringInfo();
+		copybuf = makeStringInfo();
 
-	pstate = make_parsestate(NULL);
-	(void) addRangeTableEntryForRelation(pstate, rel, AccessShareLock,
-										 NULL, false, false);
+		pstate = make_parsestate(NULL);
+		(void) addRangeTableEntryForRelation(pstate, rel, AccessShareLock,
+											 NULL, false, false);
 
-	attnamelist = make_copy_attnamelist(relmapentry);
-	cstate = BeginCopyFrom(pstate, rel, NULL, NULL, false, copy_read_data, attnamelist, options);
+		attnamelist = make_copy_attnamelist(relmapentry);
+		cstate = BeginCopyFrom(pstate, rel, NULL, NULL, false, copy_read_data, attnamelist, options);
 
-	/* Do the copy */
-	(void) CopyFrom(cstate);
+		/* Do the copy */
+		(void) CopyFrom(cstate);
+	}
 
 	logicalrep_rel_close(relmapentry, NoLock);
 }
diff --git a/src/backend/replication/pgoutput/pgoutput.c b/src/backend/replication/pgoutput/pgoutput.c
index 640676e7dc..7c54b46ac7 100644
--- a/src/backend/replication/pgoutput/pgoutput.c
+++ b/src/backend/replication/pgoutput/pgoutput.c
@@ -1462,7 +1462,6 @@ pgoutput_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	/* Switch relation if publishing via root. */
 	if (relentry->publish_as_relid != RelationGetRelid(relation))
 	{
-		Assert(relation->rd_rel->relispartition);
 		ancestor = RelationIdGetRelation(relentry->publish_as_relid);
 		targetrel = ancestor;
 	}
diff --git a/src/bin/pg_dump/t/002_pg_dump.pl b/src/bin/pg_dump/t/002_pg_dump.pl
index 0758fe5ea0..334253e489 100644
--- a/src/bin/pg_dump/t/002_pg_dump.pl
+++ b/src/bin/pg_dump/t/002_pg_dump.pl
@@ -4600,6 +4600,30 @@ my %tests = (
 			no_table_access_method => 1,
 			only_dump_measurement => 1,
 		},
+	},
+
+	'CREATE TABLE regress_pg_dump_publish_via_parent' => {
+		create_order => 3,
+		create_sql => '
+			CREATE TABLE dump_test.regress_pg_dump_publish_via_parent (col1 int);
+			CREATE TABLE dump_test.regress_pg_dump_publish_via_parent_1() INHERITS (dump_test.regress_pg_dump_publish_via_parent) WITH (publish_via_parent);',
+		regexp => qr/^
+			\QCREATE TABLE dump_test.regress_pg_dump_publish_via_parent (\E
+			\n\s+\Qcol1 integer\E
+			\n\);
+			(\n.*)+?
+			\n\QCREATE TABLE dump_test.regress_pg_dump_publish_via_parent_1 (\E
+			\n\)
+			\n\QINHERITS (dump_test.regress_pg_dump_publish_via_parent)\E
+			\n\QWITH (publish_via_parent='true')\E;/xm,
+		like => {
+			%full_runs, %dump_test_schema_runs, section_pre_data => 1,
+		},
+		unlike => {
+			binary_upgrade => 1,
+			exclude_dump_test_schema => 1,
+			only_dump_measurement => 1,
+		},
 	});
 
 #########################################
diff --git a/src/bin/psql/tab-complete.c b/src/bin/psql/tab-complete.c
index 779fdc90cb..000a996e69 100644
--- a/src/bin/psql/tab-complete.c
+++ b/src/bin/psql/tab-complete.c
@@ -1306,6 +1306,7 @@ static const char *const table_storage_parameters[] = {
 	"fillfactor",
 	"log_autovacuum_min_duration",
 	"parallel_workers",
+	"publish_via_parent",
 	"toast.autovacuum_enabled",
 	"toast.autovacuum_freeze_max_age",
 	"toast.autovacuum_freeze_min_age",
diff --git a/src/include/catalog/pg_inherits.h b/src/include/catalog/pg_inherits.h
index ce154ab943..fc539e7401 100644
--- a/src/include/catalog/pg_inherits.h
+++ b/src/include/catalog/pg_inherits.h
@@ -23,6 +23,7 @@
 
 #include "nodes/pg_list.h"
 #include "storage/lock.h"
+#include "utils/relcache.h"
 
 /* ----------------
  *		pg_inherits definition.  cpp turns this into
@@ -46,14 +47,17 @@ typedef FormData_pg_inherits *Form_pg_inherits;
 
 DECLARE_UNIQUE_INDEX_PKEY(pg_inherits_relid_seqno_index, 2680, InheritsRelidSeqnoIndexId, on pg_inherits using btree(inhrelid oid_ops, inhseqno int4_ops));
 DECLARE_INDEX(pg_inherits_parent_index, 2187, InheritsParentIndexId, on pg_inherits using btree(inhparent oid_ops));
+DECLARE_INDEX(pg_inherits_parent_seqno_index, 8138, InheritsParentSeqnoIndexId, on pg_inherits using btree(inhparent oid_ops, inhseqno int4_ops));
 
 
 extern List *find_inheritance_children(Oid parentrelId, LOCKMODE lockmode);
 extern List *find_inheritance_children_extended(Oid parentrelId, bool omit_detached,
-												LOCKMODE lockmode, bool *detached_exist, TransactionId *detached_xmin);
+												LOCKMODE lockmode, bool *detached_exist, TransactionId *detached_xmin,
+												bool logical_only);
 
 extern List *find_all_inheritors(Oid parentrelId, LOCKMODE lockmode,
 								 List **numparents);
+extern List *find_all_logical_inheritors(Oid parentrelId);
 extern bool has_subclass(Oid relationId);
 extern bool has_superclass(Oid relationId);
 extern bool typeInheritsFrom(Oid subclassTypeId, Oid superclassTypeId);
@@ -63,5 +67,7 @@ extern bool DeleteInheritsTuple(Oid inhrelid, Oid inhparent,
 								bool expect_detach_pending,
 								const char *childname);
 extern bool PartitionHasPendingDetach(Oid partoid);
+extern List *get_logical_ancestors(Oid relid, bool is_partition);
+extern bool has_logical_parent(Relation inhRel, Oid relid);
 
 #endif							/* PG_INHERITS_H */
diff --git a/src/include/catalog/pg_proc.dat b/src/include/catalog/pg_proc.dat
index 0598af0cf9..b0456ac0f2 100644
--- a/src/include/catalog/pg_proc.dat
+++ b/src/include/catalog/pg_proc.dat
@@ -11852,6 +11852,13 @@
   proname => 'pg_relation_is_publishable', provolatile => 's',
   prorettype => 'bool', proargtypes => 'regclass',
   prosrc => 'pg_relation_is_publishable' },
+{ oid => '8137',
+  descr => 'get the relations to copy from all specified publications during a table\'s initial sync',
+  proname => 'pg_get_publication_rels_to_sync', prorows => '10',
+  provariadic => 'text', proretset => 't', provolatile => 's',
+  prorettype => 'regclass', proargtypes => 'regclass _text',
+  proargmodes => '{i,v}', proargnames => '{rootid,pubnames}',
+  prosrc => 'pg_get_publication_rels_to_sync' },
 { oid => '8139',
   descr => 'get information on how a relation will be published via a list of publications',
   proname => 'pg_get_relation_publishing_info', provariadic => 'text',
diff --git a/src/include/utils/rel.h b/src/include/utils/rel.h
index 1426a353cd..3c5ea5a37a 100644
--- a/src/include/utils/rel.h
+++ b/src/include/utils/rel.h
@@ -339,6 +339,7 @@ typedef struct StdRdOptions
 	int			toast_tuple_target; /* target for tuple toasting */
 	AutoVacOpts autovacuum;		/* autovacuum-related options */
 	bool		user_catalog_table; /* use as an additional catalog relation */
+	bool		publish_via_parent; /* publish via parent's relid for logrep */
 	int			parallel_workers;	/* max number of parallel workers */
 	StdRdOptIndexCleanup vacuum_index_cleanup;	/* controls index vacuuming */
 	bool		vacuum_truncate;	/* enables vacuum to truncate a relation */
@@ -388,6 +389,17 @@ typedef struct StdRdOptions
 	  (relation)->rd_rel->relkind == RELKIND_MATVIEW) ? \
 	 ((StdRdOptions *) (relation)->rd_options)->user_catalog_table : false)
 
+/*
+ * RelationIsPublishedViaParent
+ *		Returns whether the relation's contents should be logically replicated
+ *		via its parent table's relid, when published under the effects of
+ *		publish_via_partition_root.  Note multiple eval of argument!
+ */
+#define RelationIsPublishedViaParent(relation)	\
+	((relation)->rd_options && \
+	 ((relation)->rd_rel->relkind == RELKIND_RELATION) && \
+	 ((StdRdOptions *) (relation)->rd_options)->publish_via_parent)
+
 /*
  * RelationGetParallelWorkers
  *		Returns the relation's parallel_workers reloption setting.
diff --git a/src/test/regress/expected/publication.out b/src/test/regress/expected/publication.out
index 85c217dadb..65e5cf6893 100644
--- a/src/test/regress/expected/publication.out
+++ b/src/test/regress/expected/publication.out
@@ -5,6 +5,17 @@ CREATE ROLE regress_publication_user LOGIN SUPERUSER;
 CREATE ROLE regress_publication_user2;
 CREATE ROLE regress_publication_user_dummy LOGIN NOSUPERUSER;
 SET SESSION AUTHORIZATION 'regress_publication_user';
+CREATE FUNCTION published_sync (VARIADIC pubnames text[])
+                RETURNS TABLE (published regclass, synced regclass) AS $$
+  -- Show each published table alongside the tables to be copied into it during
+  -- initial sync.
+  SELECT t.published, synced
+    FROM (SELECT DISTINCT relid::regclass AS published
+            FROM pg_get_publication_tables(VARIADIC pubnames)) t
+    JOIN LATERAL pg_get_publication_rels_to_sync(t.published, VARIADIC pubnames) synced
+      ON true
+    ORDER BY 1, 2;
+$$ LANGUAGE sql;
 CREATE FUNCTION published_stream (VARIADIC pubnames text[])
                 RETURNS TABLE (pubname text, published regclass, synced regclass) AS $$
   -- For each publication, show each published root alongside the tables which
@@ -1703,6 +1714,12 @@ SELECT * FROM pg_publication_tables;
  pub     | sch1       | tbl1      | {a}      | 
 (1 row)
 
+SELECT * FROM published_sync('pub');
+ published |  synced   
+-----------+-----------
+ sch1.tbl1 | sch1.tbl1
+(1 row)
+
 SELECT * FROM published_stream('pub');
  pubname | published |     synced      
 ---------+-----------+-----------------
@@ -1736,6 +1753,12 @@ SELECT * FROM pg_publication_tables;
  pub     | sch2       | tbl1_part1 | {a}      | 
 (1 row)
 
+SELECT * FROM published_sync('pub');
+    published    |     synced      
+-----------------+-----------------
+ sch2.tbl1_part1 | sch2.tbl1_part1
+(1 row)
+
 SELECT * FROM published_stream('pub');
  pubname |    published    |     synced      
 ---------+-----------------+-----------------
@@ -1758,12 +1781,348 @@ SELECT * FROM pg_publication_tables;
  pub     | sch1       | tbl1      | {a}      | 
 (1 row)
 
+-- Sanity check cases for publish_via_parent.
+CREATE TABLE sch1.iroot (a int);
+CREATE TABLE sch1.ipart1 (a int);
+CREATE TABLE sch1.ipart2 () INHERITS (sch1.iroot);
+-- should do nothing at all
+ALTER TABLE sch1.iroot SET (publish_via_parent = false);
+ALTER TABLE sch1.iroot RESET (publish_via_parent);
+-- marking roots between unrelated tables is not allowed
+ALTER TABLE sch1.ipart1 SET (publish_via_parent);
+ERROR:  table "ipart1" does not inherit from any tables
+CREATE TABLE fail (a int) WITH (publish_via_parent);
+ERROR:  table "fail" does not inherit from any tables
+-- establishing an inheritance relationship fixes the problem
+ALTER TABLE sch1.ipart1 INHERIT sch1.iroot,
+						SET (publish_via_parent);
+-- but multiple inheritance is not allowed
+ALTER TABLE sch1.ipart2 INHERIT sch1.ipart1,
+						SET (publish_via_parent);
+ERROR:  table "ipart2" inherits from multiple tables
+-- once publish_via_parent is set, inheritance cannot be changed
+ALTER TABLE sch1.ipart2 SET (publish_via_parent);
+ALTER TABLE sch1.ipart2 NO INHERIT sch1.iroot;
+ERROR:  relation option "publish_via_parent" prevents table "ipart2" from changing inheritance
+ALTER TABLE sch1.ipart2 INHERIT sch1.ipart1;
+ERROR:  relation option "publish_via_parent" prevents table "ipart2" from changing inheritance
+-- table ownership must match, like ATTACH PARTITION
+CREATE ROLE regress_test_me;
+CREATE ROLE regress_test_not_me;
+CREATE TABLE root (a int);
+CREATE TABLE part () INHERITS (root);
+ALTER TABLE root OWNER TO regress_test_me;
+ALTER TABLE part OWNER TO regress_test_not_me;
+SET SESSION AUTHORIZATION regress_test_me;
+ALTER TABLE part SET (publish_via_parent); -- should fail
+ERROR:  must be owner of table part
+RESET SESSION AUTHORIZATION;
+ALTER TABLE root OWNER TO regress_test_not_me;
+ALTER TABLE part OWNER TO regress_test_me;
+SET SESSION AUTHORIZATION regress_test_me;
+ALTER TABLE part SET (publish_via_parent); -- should also fail
+ERROR:  must be owner of table root
+CREATE TABLE fail () INHERITS (root) WITH (publish_via_parent);
+ERROR:  must be owner of table root
+RESET SESSION AUTHORIZATION;
+DROP TABLE root, part;
+DROP ROLE regress_test_not_me;
+DROP ROLE regress_test_me;
+-- Mixed publication settings for publish_via_partition_root, at different
+-- levels of the inheritance tree, to pin correct behavior in the worst cases.
+CREATE TABLE sch1.ipart1_a () INHERITS (sch1.ipart1) WITH (publish_via_parent);
+CREATE TABLE sch1.ipart1_a1 () INHERITS (sch1.ipart1_a);
+ALTER TABLE sch1.ipart1_a1 SET (publish_via_parent);
+CREATE PUBLICATION ipub_all FOR TABLE sch1.iroot;
+CREATE PUBLICATION ipub_one FOR TABLE ONLY sch1.ipart1_a
+       WITH (publish_via_partition_root);
+CREATE PUBLICATION ipub_two FOR TABLE ONLY sch1.ipart1_a
+       WITH (publish_via_partition_root);
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot
+--  +- ipart1
+--  | +- ipart1_a             - ipart1_a               - ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+-- A subscription to only ipub_all should see every individual table.
+SELECT relid::regclass FROM pg_get_publication_tables('ipub_all');
+     relid      
+----------------
+ sch1.iroot
+ sch1.ipart1
+ sch1.ipart2
+ sch1.ipart1_a
+ sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_sync('ipub_all');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.ipart1    | sch1.ipart1
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a  | sch1.ipart1_a
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.ipart1    | sch1.ipart1
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a  | sch1.ipart1_a
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+-- A subscription to both ipub_all and ipub_one shouldn't change the initial
+-- sync, since there is no alternative root being published for ipart1_a. It
+-- will be duplicated in the stream list, since TODO
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid      
+----------+----------------
+ ipub_all | sch1.iroot
+ ipub_all | sch1.ipart1
+ ipub_all | sch1.ipart2
+ ipub_all | sch1.ipart1_a
+ ipub_all | sch1.ipart1_a1
+ ipub_one | sch1.ipart1_a
+(6 rows)
+
+SELECT * FROM published_sync('ipub_all', 'ipub_one');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.ipart1    | sch1.ipart1
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a  | sch1.ipart1_a
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all', 'ipub_one');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.ipart1    | sch1.ipart1
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a  | sch1.ipart1_a
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+ ipub_one | sch1.ipart1_a  | sch1.ipart1_a
+(6 rows)
+
+ALTER PUBLICATION ipub_one ADD TABLE ONLY sch1.ipart1;
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot
+--  +- ipart1               - ipart1
+--  | +- ipart1_a            +- ipart1_a               - ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+-- Adding ipart1_a's parent table to ipub_one results in ipart1_a1 being
+-- "stolen" from the ipub_all stream, to prevent its data from being duplicated.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid      
+----------+----------------
+ ipub_all | sch1.iroot
+ ipub_all | sch1.ipart1
+ ipub_all | sch1.ipart2
+ ipub_all | sch1.ipart1_a1
+ ipub_one | sch1.ipart1
+(5 rows)
+
+SELECT * FROM published_sync('ipub_all', 'ipub_one');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.ipart1    | sch1.ipart1
+ sch1.ipart1    | sch1.ipart1_a
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all', 'ipub_one');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.ipart1    | sch1.ipart1
+ ipub_all | sch1.ipart1    | sch1.ipart1_a
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+ ipub_one | sch1.ipart1    | sch1.ipart1
+ ipub_one | sch1.ipart1    | sch1.ipart1_a
+(7 rows)
+
+-- Including ipub_two in the subscription list should change nothing about the
+-- sync. ipub_two will gain an entry for ipart1_a in the streaming list.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid      
+----------+----------------
+ ipub_all | sch1.iroot
+ ipub_all | sch1.ipart1
+ ipub_all | sch1.ipart2
+ ipub_all | sch1.ipart1_a1
+ ipub_one | sch1.ipart1
+(5 rows)
+
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.ipart1    | sch1.ipart1
+ sch1.ipart1    | sch1.ipart1_a
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.ipart1    | sch1.ipart1
+ ipub_all | sch1.ipart1    | sch1.ipart1_a
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+ ipub_one | sch1.ipart1    | sch1.ipart1
+ ipub_one | sch1.ipart1    | sch1.ipart1_a
+ ipub_two | sch1.ipart1    | sch1.ipart1_a
+(8 rows)
+
+ALTER PUBLICATION ipub_two ADD TABLE ONLY sch1.iroot;
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot                                          iroot
+--  +- ipart1               - ipart1               |
+--  | +- ipart1_a            +- ipart1_a           +---- ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+-- Adding iroot to ipub_two ends up stealing both ipart1 and ipart1_a from the
+-- other publications, again so that no data is double-published.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid      
+----------+----------------
+ ipub_all | sch1.iroot
+ ipub_all | sch1.ipart2
+ ipub_all | sch1.ipart1_a1
+ ipub_two | sch1.iroot
+(4 rows)
+
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.iroot     | sch1.ipart1
+ sch1.iroot     | sch1.ipart1_a
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.iroot     | sch1.ipart1
+ ipub_all | sch1.iroot     | sch1.ipart1_a
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+ ipub_one | sch1.iroot     | sch1.ipart1
+ ipub_one | sch1.iroot     | sch1.ipart1_a
+ ipub_two | sch1.iroot     | sch1.iroot
+ ipub_two | sch1.iroot     | sch1.ipart1_a
+(9 rows)
+
+-- "Detaching" partitions should change the subscriptions
+ALTER TABLE sch1.ipart1_a SET (publish_via_parent = false);
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid      
+----------+----------------
+ ipub_all | sch1.iroot
+ ipub_all | sch1.ipart2
+ ipub_all | sch1.ipart1_a
+ ipub_all | sch1.ipart1_a1
+ ipub_one | sch1.ipart1_a
+ ipub_two | sch1.iroot
+ ipub_two | sch1.ipart1_a
+(7 rows)
+
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+   published    |     synced     
+----------------+----------------
+ sch1.iroot     | sch1.iroot
+ sch1.iroot     | sch1.ipart1
+ sch1.ipart2    | sch1.ipart2
+ sch1.ipart1_a  | sch1.ipart1_a
+ sch1.ipart1_a1 | sch1.ipart1_a1
+(5 rows)
+
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+ pubname  |   published    |     synced     
+----------+----------------+----------------
+ ipub_all | sch1.iroot     | sch1.iroot
+ ipub_all | sch1.iroot     | sch1.ipart1
+ ipub_all | sch1.ipart2    | sch1.ipart2
+ ipub_all | sch1.ipart1_a  | sch1.ipart1_a
+ ipub_all | sch1.ipart1_a1 | sch1.ipart1_a1
+ ipub_one | sch1.iroot     | sch1.ipart1
+ ipub_one | sch1.ipart1_a  | sch1.ipart1_a
+ ipub_two | sch1.iroot     | sch1.iroot
+ ipub_two | sch1.ipart1_a  | sch1.ipart1_a
+(9 rows)
+
+ALTER TABLE sch1.ipart1 RESET (publish_via_parent);
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+ pubname  |     relid     
+----------+---------------
+ ipub_one | sch1.ipart1
+ ipub_one | sch1.ipart1_a
+ ipub_two | sch1.iroot
+ ipub_two | sch1.ipart1_a
+(4 rows)
+
+SELECT * FROM published_sync('ipub_one', 'ipub_two');
+   published   |    synced     
+---------------+---------------
+ sch1.iroot    | sch1.iroot
+ sch1.ipart1   | sch1.ipart1
+ sch1.ipart1_a | sch1.ipart1_a
+(3 rows)
+
+SELECT * FROM published_stream('ipub_one', 'ipub_two');
+ pubname  |   published   |    synced     
+----------+---------------+---------------
+ ipub_one | sch1.ipart1   | sch1.ipart1
+ ipub_one | sch1.ipart1_a | sch1.ipart1_a
+ ipub_two | sch1.iroot    | sch1.iroot
+ ipub_two | sch1.ipart1_a | sch1.ipart1_a
+(4 rows)
+
+DROP PUBLICATION ipub_all;
+DROP PUBLICATION ipub_one;
+DROP PUBLICATION ipub_two;
 RESET client_min_messages;
 DROP PUBLICATION pub;
 DROP TABLE sch1.tbl1;
+DROP TABLE sch1.ipart1_a1;
+DROP TABLE sch1.ipart1_a;
+DROP TABLE sch1.ipart1;
+DROP TABLE sch1.ipart2;
+DROP TABLE sch1.iroot;
 DROP SCHEMA sch1 cascade;
 DROP SCHEMA sch2 cascade;
 DROP FUNCTION published_stream;
+DROP FUNCTION published_sync;
 RESET SESSION AUTHORIZATION;
 DROP ROLE regress_publication_user, regress_publication_user2;
 DROP ROLE regress_publication_user_dummy;
diff --git a/src/test/regress/sql/publication.sql b/src/test/regress/sql/publication.sql
index 6bf9554d48..43320dfa92 100644
--- a/src/test/regress/sql/publication.sql
+++ b/src/test/regress/sql/publication.sql
@@ -6,6 +6,18 @@ CREATE ROLE regress_publication_user2;
 CREATE ROLE regress_publication_user_dummy LOGIN NOSUPERUSER;
 SET SESSION AUTHORIZATION 'regress_publication_user';
 
+CREATE FUNCTION published_sync (VARIADIC pubnames text[])
+                RETURNS TABLE (published regclass, synced regclass) AS $$
+  -- Show each published table alongside the tables to be copied into it during
+  -- initial sync.
+  SELECT t.published, synced
+    FROM (SELECT DISTINCT relid::regclass AS published
+            FROM pg_get_publication_tables(VARIADIC pubnames)) t
+    JOIN LATERAL pg_get_publication_rels_to_sync(t.published, VARIADIC pubnames) synced
+      ON true
+    ORDER BY 1, 2;
+$$ LANGUAGE sql;
+
 CREATE FUNCTION published_stream (VARIADIC pubnames text[])
                 RETURNS TABLE (pubname text, published regclass, synced regclass) AS $$
   -- For each publication, show each published root alongside the tables which
@@ -1076,6 +1088,7 @@ SELECT * FROM pg_publication_tables;
 -- Table publication that includes both the parent table and the child table
 ALTER PUBLICATION pub ADD TABLE sch1.tbl1;
 SELECT * FROM pg_publication_tables;
+SELECT * FROM published_sync('pub');
 SELECT * FROM published_stream('pub');
 
 DROP PUBLICATION pub;
@@ -1091,6 +1104,7 @@ SELECT * FROM pg_publication_tables;
 -- Table publication that includes both the parent table and the child table
 ALTER PUBLICATION pub ADD TABLE sch1.tbl1;
 SELECT * FROM pg_publication_tables;
+SELECT * FROM published_sync('pub');
 SELECT * FROM published_stream('pub');
 
 DROP PUBLICATION pub;
@@ -1105,12 +1119,164 @@ ALTER TABLE sch1.tbl1 ATTACH PARTITION sch1.tbl1_part3 FOR VALUES FROM (20) to (
 CREATE PUBLICATION pub FOR TABLES IN SCHEMA sch1 WITH (PUBLISH_VIA_PARTITION_ROOT=1);
 SELECT * FROM pg_publication_tables;
 
+-- Sanity check cases for publish_via_parent.
+CREATE TABLE sch1.iroot (a int);
+CREATE TABLE sch1.ipart1 (a int);
+CREATE TABLE sch1.ipart2 () INHERITS (sch1.iroot);
+
+-- should do nothing at all
+ALTER TABLE sch1.iroot SET (publish_via_parent = false);
+ALTER TABLE sch1.iroot RESET (publish_via_parent);
+
+-- marking roots between unrelated tables is not allowed
+ALTER TABLE sch1.ipart1 SET (publish_via_parent);
+CREATE TABLE fail (a int) WITH (publish_via_parent);
+
+-- establishing an inheritance relationship fixes the problem
+ALTER TABLE sch1.ipart1 INHERIT sch1.iroot,
+						SET (publish_via_parent);
+
+-- but multiple inheritance is not allowed
+ALTER TABLE sch1.ipart2 INHERIT sch1.ipart1,
+						SET (publish_via_parent);
+
+-- once publish_via_parent is set, inheritance cannot be changed
+ALTER TABLE sch1.ipart2 SET (publish_via_parent);
+ALTER TABLE sch1.ipart2 NO INHERIT sch1.iroot;
+ALTER TABLE sch1.ipart2 INHERIT sch1.ipart1;
+
+-- table ownership must match, like ATTACH PARTITION
+CREATE ROLE regress_test_me;
+CREATE ROLE regress_test_not_me;
+CREATE TABLE root (a int);
+CREATE TABLE part () INHERITS (root);
+
+ALTER TABLE root OWNER TO regress_test_me;
+ALTER TABLE part OWNER TO regress_test_not_me;
+SET SESSION AUTHORIZATION regress_test_me;
+ALTER TABLE part SET (publish_via_parent); -- should fail
+RESET SESSION AUTHORIZATION;
+
+ALTER TABLE root OWNER TO regress_test_not_me;
+ALTER TABLE part OWNER TO regress_test_me;
+SET SESSION AUTHORIZATION regress_test_me;
+ALTER TABLE part SET (publish_via_parent); -- should also fail
+CREATE TABLE fail () INHERITS (root) WITH (publish_via_parent);
+RESET SESSION AUTHORIZATION;
+
+DROP TABLE root, part;
+DROP ROLE regress_test_not_me;
+DROP ROLE regress_test_me;
+
+-- Mixed publication settings for publish_via_partition_root, at different
+-- levels of the inheritance tree, to pin correct behavior in the worst cases.
+CREATE TABLE sch1.ipart1_a () INHERITS (sch1.ipart1) WITH (publish_via_parent);
+CREATE TABLE sch1.ipart1_a1 () INHERITS (sch1.ipart1_a);
+ALTER TABLE sch1.ipart1_a1 SET (publish_via_parent);
+
+CREATE PUBLICATION ipub_all FOR TABLE sch1.iroot;
+CREATE PUBLICATION ipub_one FOR TABLE ONLY sch1.ipart1_a
+       WITH (publish_via_partition_root);
+CREATE PUBLICATION ipub_two FOR TABLE ONLY sch1.ipart1_a
+       WITH (publish_via_partition_root);
+
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot
+--  +- ipart1
+--  | +- ipart1_a             - ipart1_a               - ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+
+-- A subscription to only ipub_all should see every individual table.
+SELECT relid::regclass FROM pg_get_publication_tables('ipub_all');
+SELECT * FROM published_sync('ipub_all');
+SELECT * FROM published_stream('ipub_all');
+
+-- A subscription to both ipub_all and ipub_one shouldn't change the initial
+-- sync, since there is no alternative root being published for ipart1_a. It
+-- will be duplicated in the stream list, since TODO
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_all', 'ipub_one');
+SELECT * FROM published_stream('ipub_all', 'ipub_one');
+
+ALTER PUBLICATION ipub_one ADD TABLE ONLY sch1.ipart1;
+
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot
+--  +- ipart1               - ipart1
+--  | +- ipart1_a            +- ipart1_a               - ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+
+-- Adding ipart1_a's parent table to ipub_one results in ipart1_a1 being
+-- "stolen" from the ipub_all stream, to prevent its data from being duplicated.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_all', 'ipub_one');
+SELECT * FROM published_stream('ipub_all', 'ipub_one');
+
+-- Including ipub_two in the subscription list should change nothing about the
+-- sync. ipub_two will gain an entry for ipart1_a in the streaming list.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+
+ALTER PUBLICATION ipub_two ADD TABLE ONLY sch1.iroot;
+
+-- ipub_all              ipub_one (pubviaroot)    ipub_two (pubviaroot)
+-- ------------------    ---------------------    ---------------------
+--  iroot                                          iroot
+--  +- ipart1               - ipart1               |
+--  | +- ipart1_a            +- ipart1_a           +---- ipart1_a
+--  |   +- ipart1_a1
+--  +- ipart2
+
+-- Adding iroot to ipub_two ends up stealing both ipart1 and ipart1_a from the
+-- other publications, again so that no data is double-published.
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+
+-- "Detaching" partitions should change the subscriptions
+ALTER TABLE sch1.ipart1_a SET (publish_via_parent = false);
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_all', 'ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_all', 'ipub_one', 'ipub_two');
+SELECT * FROM published_stream('ipub_all', 'ipub_one', 'ipub_two');
+
+ALTER TABLE sch1.ipart1 RESET (publish_via_parent);
+SELECT pubname, relid::regclass
+  FROM pg_get_publication_tables('ipub_one', 'ipub_two') t
+  JOIN pg_publication p ON (p.oid = t.pubid);
+SELECT * FROM published_sync('ipub_one', 'ipub_two');
+SELECT * FROM published_stream('ipub_one', 'ipub_two');
+
+DROP PUBLICATION ipub_all;
+DROP PUBLICATION ipub_one;
+DROP PUBLICATION ipub_two;
+
 RESET client_min_messages;
 DROP PUBLICATION pub;
 DROP TABLE sch1.tbl1;
+DROP TABLE sch1.ipart1_a1;
+DROP TABLE sch1.ipart1_a;
+DROP TABLE sch1.ipart1;
+DROP TABLE sch1.ipart2;
+DROP TABLE sch1.iroot;
 DROP SCHEMA sch1 cascade;
 DROP SCHEMA sch2 cascade;
 DROP FUNCTION published_stream;
+DROP FUNCTION published_sync;
 
 RESET SESSION AUTHORIZATION;
 DROP ROLE regress_publication_user, regress_publication_user2;
diff --git a/src/test/subscription/t/013_partition.pl b/src/test/subscription/t/013_partition.pl
index 275fb3b525..09e3f2d028 100644
--- a/src/test/subscription/t/013_partition.pl
+++ b/src/test/subscription/t/013_partition.pl
@@ -388,6 +388,59 @@ $node_subscriber1->append_conf('postgresql.conf',
 	"log_min_messages = warning");
 $node_subscriber1->reload;
 
+# Make sure standard inheritance setups aren't broken by the new
+# publish_via_parent handling.
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1 (a int, b text)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1_1 (LIKE itab1)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1_2 (LIKE itab1)");
+
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1 (a int, b text)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1_1 (CHECK (a = 1)) INHERITS (itab1)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1_2 (CHECK (a = 2)) INHERITS (itab1)");
+
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab1_1 SET (publish_via_parent = true)");
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab1_2 SET (publish_via_parent = true)");
+
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1 VALUES (0, 'itab1')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1_1 VALUES (1, 'itab1')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1_2 VALUES (2, 'itab1')");
+
+# Regression: Create a publication for an unrelated table and set
+# publish_via_partition_root. This should have no effect at all, since itab1
+# isn't supposed to be using this publication...
+$node_publisher->safe_psql('postgres',
+	"CREATE PUBLICATION dummy_pub FOR TABLE tab1 WITH (publish_via_partition_root = true)");
+$node_subscriber2->safe_psql('postgres',
+	"ALTER SUBSCRIPTION sub2 ADD PUBLICATION dummy_pub");
+$node_subscriber2->wait_for_subscription_sync;
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1");
+is($result, qq(0|itab1), 'initial data synced for itab1 on subscriber 2');
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1_1");
+is($result, qq(1|itab1), 'initial data synced for itab1_1 on subscriber 2');
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1_2");
+is($result, qq(2|itab1), 'initial data synced for itab1_2 on subscriber 2');
+
+$node_publisher->safe_psql('postgres',
+	"DROP TABLE itab1 CASCADE;");
+$node_subscriber2->safe_psql('postgres',
+	"DROP TABLE itab1 CASCADE;");
+
+$node_subscriber2->safe_psql('postgres',
+	"ALTER SUBSCRIPTION sub2 DROP PUBLICATION dummy_pub");
+$node_publisher->safe_psql('postgres', "DROP PUBLICATION dummy_pub");
+
 # Tests for replication using root table identity and schema
 
 # publisher
@@ -886,4 +939,263 @@ $result = $node_subscriber2->safe_psql('postgres',
 	"SELECT a, b, c FROM tab5_1 ORDER BY 1");
 is($result, qq(4||1), 'updates of tab5 replicated correctly');
 
+# Test that replication works for older inheritance/trigger setups as well.
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1 (a int, b text)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1_1 (CHECK (a = 1)) INHERITS (itab1)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab1_2 (CHECK (a = 2)) INHERITS (itab1)");
+
+$node_publisher->safe_psql('postgres', "
+	CREATE OR REPLACE FUNCTION itab1_trigger()
+	RETURNS TRIGGER AS \$\$
+	BEGIN
+		IF ( NEW.a = 1 ) THEN INSERT INTO itab1_1 VALUES (NEW.*);
+		ELSIF ( NEW.a = 2 ) THEN INSERT INTO itab1_2 VALUES (NEW.*);
+		ELSE RETURN NEW;
+		END IF;
+		RETURN NULL;
+	END;
+	\$\$
+	LANGUAGE plpgsql;");
+$node_publisher->safe_psql('postgres', "
+	CREATE TRIGGER itab1_trigger
+	BEFORE INSERT ON itab1
+	FOR EACH ROW EXECUTE FUNCTION itab1_trigger();");
+
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab1_1 SET (publish_via_parent = true)");
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab1_2 SET (publish_via_parent = true)");
+
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab2 (a int, b text)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab2_1 (CHECK (a = 1)) INHERITS (itab2)");
+
+$node_publisher->safe_psql('postgres', "
+	CREATE OR REPLACE FUNCTION itab2_trigger()
+	RETURNS TRIGGER AS \$\$
+	BEGIN
+		IF ( NEW.a = 1 ) THEN INSERT INTO itab2_1 VALUES (NEW.*);
+		ELSE RETURN NEW;
+		END IF;
+		RETURN NULL;
+	END;
+	\$\$
+	LANGUAGE plpgsql;");
+$node_publisher->safe_psql('postgres', "
+	CREATE TRIGGER itab2_trigger
+	BEFORE INSERT ON itab2
+	FOR EACH ROW EXECUTE FUNCTION itab2_trigger();");
+
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab2_1 SET (publish_via_parent = true)");
+
+# itab2_1 should be published using its own identity here, since its parent is
+# not included. itab1_1 should be published via its parent, itab1, without
+# duplicating the rows.
+$node_publisher->safe_psql('postgres',
+	"ALTER PUBLICATION pub_viaroot ADD TABLE itab1, itab1_1, itab2_1");
+
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1 VALUES (0, 'itab1')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1 VALUES (1, 'itab1')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1 VALUES (2, 'itab1')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab2 VALUES (0, 'itab2')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab2 VALUES (1, 'itab2')");
+
+# Subscriber 1 only subscribes to some of the partitions, and does not set up
+# partition triggers, to check for the correct routing.
+$node_subscriber1->safe_psql('postgres',
+	"CREATE TABLE itab1 (a int, b text)");
+$node_subscriber1->safe_psql('postgres',
+	"CREATE TABLE itab1_1 (CHECK (a = 1)) INHERITS (itab1)");
+$node_subscriber1->safe_psql('postgres',
+	"CREATE TABLE itab2_1 (a int, b text)");
+
+# Subscriber 2 has different partition names for itab1, and it doesn't partition
+# itab2 at all.
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1 (a int, b text)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1_part1 (CHECK (a = 1)) INHERITS (itab1)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab1_part2 (CHECK (a = 2)) INHERITS (itab1)");
+
+$node_subscriber2->safe_psql('postgres', "
+	CREATE OR REPLACE FUNCTION itab_trigger()
+	RETURNS TRIGGER AS \$\$
+	BEGIN
+		IF ( NEW.a = 1 ) THEN INSERT INTO public.itab1_part1 VALUES (NEW.*);
+		ELSIF ( NEW.a = 2 ) THEN INSERT INTO public.itab1_part2 VALUES (NEW.*);
+		ELSE RETURN NEW;
+		END IF;
+		RETURN NULL;
+	END;
+	\$\$
+	LANGUAGE plpgsql;");
+$node_subscriber2->safe_psql('postgres', "
+	CREATE TRIGGER itab_trigger
+	BEFORE INSERT ON itab1
+	FOR EACH ROW EXECUTE FUNCTION itab_trigger();");
+$node_subscriber2->safe_psql('postgres', "
+	ALTER TABLE itab1 ENABLE ALWAYS TRIGGER itab_trigger;");
+
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab2 (a int, b text)");
+
+$node_subscriber1->safe_psql('postgres',
+	"ALTER SUBSCRIPTION sub_viaroot REFRESH PUBLICATION");
+$node_subscriber2->safe_psql('postgres',
+	"ALTER SUBSCRIPTION sub2 REFRESH PUBLICATION");
+
+$node_subscriber1->wait_for_subscription_sync;
+$node_subscriber2->wait_for_subscription_sync;
+
+# check that data is synced correctly
+
+$result = $node_subscriber1->safe_psql('postgres',
+	"SELECT a, b FROM itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1
+1|itab1
+2|itab1), 'initial data synced for itab1 on subscriber 1');
+
+# all of the data should have been routed to itab1 directly (there are no
+# triggers on subscriber 1 to move it elsewhere)
+$result = $node_subscriber1->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1
+1|itab1
+2|itab1), 'initial data correctly routed for itab1 on subscriber 1');
+
+$result = $node_subscriber1->safe_psql('postgres',
+	"SELECT a, b FROM itab2_1 ORDER BY 1, 2");
+is($result, qq(1|itab2), 'initial data synced for itab2_1 on subscriber 1');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1
+1|itab1
+2|itab1), 'initial data synced for itab1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1), 'initial data correctly routed for itab1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab2 ORDER BY 1, 2");
+is($result, qq(0|itab2
+1|itab2), 'initial data synced for itab2 on subscriber 2');
+
+# make sure new data is also correctly routed to the roots
+$node_publisher->safe_psql('postgres', "INSERT INTO itab1 VALUES (1, 'itab1-new')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab2 VALUES (1, 'itab2-new')");
+
+$node_publisher->wait_for_catchup('sub_viaroot');
+$node_publisher->wait_for_catchup('sub2');
+
+$result = $node_subscriber1->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1
+1|itab1
+1|itab1-new
+2|itab1), 'new data routed for itab1 on subscriber 1');
+
+$result = $node_subscriber1->safe_psql('postgres',
+	"SELECT a, b FROM itab2_1 ORDER BY 1, 2");
+is($result, qq(1|itab2
+1|itab2-new), 'new data routed for itab2_1 on subscriber 1');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1 ORDER BY 1, 2");
+is($result, qq(0|itab1
+1|itab1
+1|itab1-new
+2|itab1), 'new data routed for itab1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab1_part1 ORDER BY 1, 2");
+is($result, qq(1|itab1
+1|itab1-new), 'new data moved to itab1_part1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM itab2 ORDER BY 1, 2");
+is($result, qq(0|itab2
+1|itab2
+1|itab2-new), 'new data routed for itab2 on subscriber 2');
+
+# Finally, check the effect of mixed publish_via_partition_root settings on
+# logical roots.
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab3 (a int, b text)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab3_1 () INHERITS (itab3)");
+$node_publisher->safe_psql('postgres',
+	"CREATE TABLE itab3_1_1 () INHERITS (itab3_1)");
+
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab3_1 SET (publish_via_parent = true)");
+$node_publisher->safe_psql('postgres',
+	"ALTER TABLE itab3_1_1 SET (publish_via_parent = true)");
+
+$node_publisher->safe_psql('postgres',
+	"INSERT INTO itab3 VALUES (1, 'itab3')");
+$node_publisher->safe_psql('postgres',
+	"INSERT INTO itab3_1 VALUES (2, 'itab3_1')");
+$node_publisher->safe_psql('postgres',
+	"INSERT INTO itab3_1_1 VALUES (3, 'itab3_1_1')");
+
+$node_publisher->safe_psql('postgres',
+	"CREATE PUBLICATION ipub3 FOR TABLE itab3 *");
+$node_publisher->safe_psql('postgres',
+	"CREATE PUBLICATION ipub3_1 FOR TABLE itab3_1 * WITH (publish_via_partition_root = true)");
+
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab3 (a int, b text)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab3_1 () INHERITS (itab3)");
+$node_subscriber2->safe_psql('postgres',
+	"CREATE TABLE itab3_1_1 () INHERITS (itab3_1)");
+
+$node_subscriber2->safe_psql('postgres',
+	"CREATE SUBSCRIPTION mixed CONNECTION '$publisher_connstr' PUBLICATION ipub3, ipub3_1");
+$node_subscriber2->wait_for_subscription_sync;
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3");
+is($result, qq(1|itab3), 'initial data routed for itab3 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3_1 ORDER BY 1, 2");
+is($result, qq(2|itab3_1
+3|itab3_1_1), 'initial data routed for itab3_1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3_1_1");
+is($result, qq(), 'initial data routed for itab3_1_1 on subscriber 2');
+
+# make sure new data is also correctly routed to the roots
+$node_publisher->safe_psql('postgres', "INSERT INTO itab3 VALUES (4, 'itab3_new')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab3_1 VALUES (4, 'itab3_1_new')");
+$node_publisher->safe_psql('postgres', "INSERT INTO itab3_1_1 VALUES (4, 'itab3_1_1_new')");
+
+$node_publisher->wait_for_catchup('mixed');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3 ORDER BY 1, 2");
+is($result, qq(1|itab3
+4|itab3_new), 'new data routed for itab3 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3_1 ORDER BY 1, 2");
+is($result, qq(2|itab3_1
+3|itab3_1_1
+4|itab3_1_1_new
+4|itab3_1_new), 'new data routed for itab3_1 on subscriber 2');
+
+$result = $node_subscriber2->safe_psql('postgres',
+	"SELECT a, b FROM ONLY itab3_1_1");
+is($result, qq(), 'new data routed for itab3_1_1 on subscriber 2');
+
 done_testing();
-- 
2.34.1

